import hgtiles.format as hgfo
import pandas as pd

import numpy as np
import pandas as pd
import h5py

def csv_to_points(csv_file, output_file):
    '''
    Convert a csv file containing points to a numpy array 
    of [[x,y]] values.

    Parameters:
    -----------
    csv_file: string 
        The filename of the data file 

    '''
    df = pd.read_table(csv_file, delimiter=',')

    print("dataframe", df.head(), df.columns.values)
    
    min_x = df['x'].min()
    max_x = df['x'].max()
    min_y = df['y'].min()
    max_y = df['y'].max()
    
    width = max_x - min_x
    height = max_y - min_y
    
    max_width = min(width, height)
    max_zoom = 30
    
    with h5py.File(output_file, 'w') as f_out:
        dataset = f_out.create_dataset('values', (len(df), 2), compression='gzip', dtype=np.float32)
        dataset[:] = df.reindex(columns=['x','y']).as_matrix()
        
        dataset.attrs['min_x'] = min_x
        dataset.attrs['max_x'] = max_x
        dataset.attrs['min_y'] = min_y
        dataset.attrs['max_y'] = max_y
        dataset.attrs['max_zoom'] = max_zoom
        dataset.attrs['max_width'] = max_width

    
    info = {
        'min_pos': [min_x, min_y],
        'max_pos': [min_y, max_y],
        'max_zoom': max_zoom,
        'max_width': max_width
    }
    
    return df.reindex(columns=['x', 'y'])

def tileset_info(points_file):
    '''
    Calculate the extent, etc...
    '''
    print("points file:", points_file)

    with h5py.File(points_file, 'r') as f_in:
        dset = f_in['values']
        
        return {
            'min_pos': [float(dset.attrs['min_x']), float(dset.attrs['min_y'])],
            'max_pos': [float(dset.attrs['max_y']), float(dset.attrs['max_y'])],
            'max_width': float(dset.attrs['max_width']),
            'max_zoom': int(dset.attrs['max_zoom']),
            'mirror_tiles': 'false'
        }

def tile_bounds(points_file, z, x, y):
    '''
    Get the boundaries of a tile
    
    Parameters:
    -----------
    tileset_info: { min_pos, max_pos, max_width}
        Information about the bounds of this tileset
    '''
    tsinfo = tileset_info(points_file)
    tile_width = tsinfo['max_width'] / 2 ** z
    print("tile_width:", tile_width)

        
    x_start = tsinfo['min_pos'][0] + tile_width * x
    x_end = tsinfo['min_pos'][0] + tile_width * (x+1)
    y_start = tsinfo['min_pos'][1] + tile_width * y
    y_end = tsinfo['min_pos'][1] + tile_width * (y+1)
    
    return (x_start, x_end, y_start, y_end)
    
def filter_points(data, extent):
    '''
    Filter points that are within the extent
    
    Parameters:
    -----------
    data: [[]]
        A 2D numpy array containing x,y values
        
    extent: [x_start, x_end, y_start, y_end]
        The region we want to return points within
        
    Returns
    -------
    data: [[]]
        A 2D numpy array containing x,y values
    '''
    # print("extent:", extent)
    # print("data.shape", data.shape, data[:,0])

    print("data:", data[:10])

    data = data[data[:,0] > extent[0]]
    data = data[data[:,0] < extent[1]]
    
    data = data[data[:,1] > extent[2]]
    data = data[data[:,1] < extent[3]]
        
    print("extent:", extent, "len(data):", len(data))

    return data

def density_tiles(points_file, z, x, y):
    with h5py.File(points_file, 'r') as f:
        filtered_points = filter_points(f['values'][:], 
                                        tile_bounds(points_file, z, x, y))
        dt = np.histogram2d(filtered_points[:,0], 
                              filtered_points[:,1], bins=256)[0].T
        dt[dt  == 0.] = np.nan
        return dt

def tiles(points_file, z, x, y):
    return hgfo.format_dense_tile(
        density_tiles(points_file, z, x, y).flatten(),
    )



import base64
import numpy as np

def format_dense_tile(data):
    '''
    Format raw vector data into a more structured tile
    containing either float16 or float32 data along with a 
    dtype to differentiate between the two.

    Parameters
    ----------
    tile_data_array: np.array
        An array of values
    Returns
    -------
    tile_data: {'dense': str, 'dtype': str}
        The tile data reformatted to use float16 or float32 as the
        datatype. The dtype indicates which format is chosen.
    '''

    tile_data = {}

    if len(data):
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')

            max_dense = float(np.nanmax(data))
            min_dense = float(np.nanmin(data))
    else:
        max_dense = np.nan
        min_dense = np.nan

    tile_data["min_value"] = min_dense if not np.isnan(min_dense) else "NaN"
    tile_data["max_value"] = max_dense if not np.isnan(max_dense) else "NaN"

    min_f16 = np.finfo('float16').min
    max_f16 = np.finfo('float16').max

    has_nan = len([d for d in data if np.isnan(d)]) > 0

    if (
        not has_nan and
        max_dense > min_f16 and max_dense < max_f16 and
        min_dense > min_f16 and min_dense < max_f16
    ):
        tile_data = {
            'dense': base64.b64encode(data.astype('float16')).decode('utf-8'),
            'dtype': 'float16'
        }
    else:
        tile_data = {
            'dense': base64.b64encode(data.astype('float32')).decode('utf-8'),
            'dtype': 'float32'
        }

    return tile_data




from setuptools import setup
from setuptools.command.build_ext import build_ext as _build_ext

install_requires = [
    'cython',
    'numpy',
    'pysam',
    'h5py',
    'pandas',
    'slugid',
    'numpy',
    'scipy==1.0.1',
    'cooler',
    'pybbi==0.1.3',
]

setup(
    name='hgtiles',
    version='0.2.13',
    description='Tile loading for higlass-server',
    author='Peter Kerpedjiev',
    author_email='pkerpedjiev@gmail.com',
    url='',
    install_requires=install_requires,
    setup_requires=['numpy'],
    packages=['hgtiles']
)



import collections as col
import json
import os
import os.path as op
import random
import sqlite3
import sys

def csv_to_hg(csv_file, output_file, delimiter=','):
    conn = sqlite3.connect(output_file)

    conn.execute('''
    CREATE TABLE data_info (
        minX NUMBER,
        maxX NUMBER,
        minY NUMBER,
        maxY NUMBER,
        maxZoom INT,
        numPoints INT,
        maxWidth NUMBER )
    ''')

    conn.execute('''
        CREATE TABLE points
        (
            id int PRIMARY KEY,
            x NUMBER,
            y NUMBER,
            importance NUMBER,
            data TEXT
        )
        ''')
    
    conn.execute('''
    CREATE INDEX points_importance ON points ( importance )
    ''')
    
    conn.execute('''
        CREATE VIRTUAL TABLE points_index USING rtree(
            id, 
            minX, maxX,
            minY, maxY,
            minZ, maxZ
        )
        ''')

    min_x, min_y = sys.maxsize, sys.maxsize
    max_x, max_y = -sys.maxsize, -sys.maxsize
    points = []
    
    # go through the file once to calculate the extent
    with open(csv_file, 'r') as f:
        # skip the header
        f.readline()
        
        for i,line in enumerate(f):
            parts = line.strip().split(delimiter)
            #print("parts:", parts)
            x = float(parts[0])
            y = float(parts[1])
            point = {
                'uid': i,
                'pos': [x, y],
                'importance': random.random(),
                'data': parts[2]
            }
            
            min_x = x if x < min_x else min_x
            max_x = x if x > max_x else max_x
            min_y = y if y < min_y else min_y
            max_y = y if y > max_y else max_y
            
            points += [point]
            
    extent = [min_x, min_y, max_x, max_y]
    
    width = extent[2] - extent[0]
    height = extent[3] - extent[1]
    
    max_width = max(width, height)
    max_zoom = 20
    max_per_tile = 100
    tile_counts = col.defaultdict(int)
    
    # go through and assign zoom levels to each point

    counter = 0
    sorted_points = sorted(points, key=lambda x: -x['importance'])
    for point in sorted_points:
        curr_zoom = 0
        
        while curr_zoom <= max_zoom:
            tile_width = max_width / 2 ** curr_zoom
            x,y = point['pos']
            
            tile_x, tile_y = (int((x - min_x) // tile_width), 
                              int((y - min_y) // tile_width))
            #print("tile_x:", tile_x, "tile_y:", tile_y)

            
            if tile_counts[(curr_zoom, tile_x, tile_y)] < max_per_tile:
                # there's space in this tile, so add this point
                tile_counts[(curr_zoom, tile_x, tile_y)] += 1
                point['zoom'] = curr_zoom
                conn.execute(
                '''
                INSERT INTO points
                VALUES (?,?,?,?,?)
                ''', (
                    counter,
                    point['pos'][0],
                    point['pos'][1],
                    point['importance'],
                    json.dumps(point)
                ))
                
                #if curr_zoom == 2 and tile_x == 0 and tile_y == 1:
                #    print("hi", i, curr_zoom, x, y, tile_x, tile_y)                
                conn.execute(
                'INSERT INTO points_index VALUES(?,?,?,?,?,?,?)',
                (counter, x, x, y, y, counter, counter))
                
                counter += 1
                break

            # not enough space so go onto the next zoom level
            curr_zoom += 1
                
    max_zoom = 30
    num_points = counter 

    conn.execute('''
    INSERT INTO data_info
    VALUES (?,?,?,?,?,?,?)
    ''',
    (extent[0], extent[1],
        extent[2], extent[3],
        max_zoom, num_points, max_width))

    conn.commit()
    conn.close()
    
    return {
        "data": sorted_points,
        "conn": conn,
        "extent": extent,
        "min_pos": [extent[0], extent[1]],
        "max_pos": [extent[2], extent[3]],
        "max_width": max_width,
        "max_zoom": max_zoom, # arbitrary large value
        "num_points": counter
    }
    return conn

def tileset_info(db_file):
    conn = sqlite3.connect(db_file)
    
    row = conn.execute('''
        SELECT * from data_info
        ''').fetchone()

    return {
        'min_pos': [row[0], row[1]],
        'max_pos': [row[2], row[3]],
        'max_zoom': row[4],
        'num_points': row[5],
        'max_width': row[6] 
    }
    

def tiles(filename, z, x, y, max_points=40):
    hg_points = tileset_info(filename)
    conn = sqlite3.connect(filename)

    area_width = hg_points['max_width'] / 2 ** z
    extent = hg_points['min_pos'] + hg_points['max_pos']
    
    min_x = extent[0] + area_width * x
    max_x = extent[0] + area_width * (x+1)
    min_y = extent[1] + area_width * y
    max_y = extent[1] + area_width * (y+1)
    min_z = -1
    
    
    print("minx", min_x, min_y, max_x, max_y)

    points_to_return = []
    
    while len(points_to_return) < max_points and z < hg_points['max_zoom']:
        max_z = 30 * 2 ** (z+1)

        #print("max_z:", max_z)

        query = '''
        SELECT x,y,importance,data 
        FROM points, points_index
        WHERE 
            points.id = points_index.id AND
            maxX >= {} AND
            maxY >= {} AND
            maxZ >= {} AND
            minX < {} AND
            minY < {} AND
            minZ < {}
        ORDER BY
            importance
        DESC
        LIMIT {}'''.format(min_x, min_y, min_z, max_x, max_y, max_z, max_points)
        #print("query:", query)

        c = conn.cursor()
        points_to_return = [json.loads(r[3]) for r in c.execute(query).fetchall()]
        #print("len:", len(points_to_return))
        #print("z:", z)
        z += 1
        z *= 2
        
    return points_to_return
